#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <signal.h>

#ifdef USE_GRAPHIC
#ifdef WIN32
#ifdef USE_GRAPHIC_SDL1
#include <SDL/SDL.h>
#endif
#ifdef USE_GRAPHIC_SDL2
#include <SDL2/SDL.h>
#endif
#endif
#endif

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "types.h"
#include "nlllib.h"
#include "readline.h"
#include "argument.h"
#include "nll.h"

static void signal_init(void);

static void sigint_handler(int arg)
{
  nll_finish();
#ifdef WIN32
  signal_init();
#endif
  return;
}

#ifdef WIN32
#define USE_OLD_SIGNAL
#endif

static void signal_init(void)
{
#ifndef USE_OLD_SIGNAL
  struct sigaction sa;

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = sigint_handler;

  sigaction(SIGINT, &sa, NULL);
#else
  signal(SIGINT, sigint_handler);
#endif
}

static void signal_reset(void)
{
#ifndef USE_OLD_SIGNAL
  struct sigaction sa;

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = SIG_DFL;

  sigaction(SIGINT, &sa, NULL);
#else
  signal(SIGINT, SIG_DFL);
#endif
}

static void init(void)
{
  int r;

  signal_init();

  if ((r = nll_init(stdin, stdout)) < 0)
    NLL_ERREXIT(r, "(EXIT)");
}

static void done(void)
{
  int r;

  if ((r = nll_done()) < 0)
    NLL_ERREXIT(r, "(EXIT)");

  signal_reset();
}

static void truncate_line(char *buffer)
{
  int len;
  if (buffer) {
    len = strlen(buffer);
    while (len > 0) {
      if (!isspace(buffer[len - 1]))
	break;
      len--;
    }
    buffer[len] = '\0';
  }
}

static void help(void)
{
  printf("nll [-help|-h] [-quit|-q] [-no-run|-nr] [-number|-n] [-e <command>] <file>\n");
  printf("\t-help,-h\tOutput help message.\n");
  printf("\t-quit,-q\tAuto quit after execution of the input file.\n");
  printf("\t-no-run,-nr\tDisable auto execution of the input file.\n");
  printf("\t-number,-n\tAuto numbering mode of the input file.\n");
  printf("\t-e <command>\tExecute the command.\n");
  printf("Samples:\n");
  printf("\t$ nll sample.hob\n");
  printf("\t$ echo \"abc\" | nll -e \'S=FGETS(0);FPUTS(0,S)\'\n");
  exit(0);
}

static int quit        = 0;
static int auto_run    = 1;
static int auto_number = 0;
static int no_system   = 0;
static int no_syscall  = 0;
static char *command = NULL;

static argument_t args[] = {
  { "-help"      , ARGUMENT_TYPE_FUNCTION, help         },
  { "-h"         , ARGUMENT_TYPE_FUNCTION, help         },
  { "-quit"      , ARGUMENT_TYPE_FLAG_ON , &quit        },
  { "-q"         , ARGUMENT_TYPE_FLAG_ON , &quit        },
  { "-no-run"    , ARGUMENT_TYPE_FLAG_OFF, &auto_run    },
  { "-nr"        , ARGUMENT_TYPE_FLAG_OFF, &auto_run    },
  { "-number"    , ARGUMENT_TYPE_FLAG_ON , &auto_number },
  { "-n"         , ARGUMENT_TYPE_FLAG_ON , &auto_number },
  { "-no-system" , ARGUMENT_TYPE_FLAG_ON , &no_system   },
  { "-ns"        , ARGUMENT_TYPE_FLAG_ON , &no_system   },
  { "-no-syscall", ARGUMENT_TYPE_FLAG_ON , &no_syscall  },
  { "-nc"        , ARGUMENT_TYPE_FLAG_ON , &no_syscall  },
  { "-e"         , ARGUMENT_TYPE_STRING  , &command     },
  { NULL, ARGUMENT_TYPE_NONE, NULL }
};

int main(int argc, char *argv[])
{
  char *filename = NULL, *linebuf, *readlinebuf, *top, *p, *head;
  int finished = 0, initialize, r, number, number2, retcode = 0;
  line_t line;
  FILE *fp = NULL;
  char buffer[LINE_MAXLEN + 1 + 1]; /* +1 is for auto number space */

  argument_read(&argc, argv, args);

  argc--;
  argv++;

  if (no_system)
    nll_set_nosystem();
  if (no_syscall)
    nll_set_nosyscall();

  if (argc > 0) {
    filename = argv[0];
  }

  init();

  if (filename) {
    if ((r = nll_file_open(filename)) < 0)
      NLL_ERREXIT(r, filename);
  }

  nll_istty = isatty(STDIN_FILENO);

  while (!finished) {
    fp = nll_file_fp();
    if (!fp)
      auto_number = 0;

    linebuf = NULL;
    readlinebuf = NULL;

    if (fp) {
      linebuf = buffer + (auto_number ? 1 : 0);
      if (!fgets(linebuf, LINE_MAXLEN + 1, fp)) {
	nll_file_close();
	continue;
      }
      truncate_line(linebuf);
    } else if (command) {
      linebuf = command;
      command = NULL;
    } else if (auto_run) {
      linebuf = "RUN";
      auto_run = 0;
    } else if (quit) {
      finished = 1;
      continue;
    } else if (nll_istty) {
      readlinebuf = nll_readline("nll> ");
      if (!readlinebuf) {
	fprintf(stdout, "\n");
	continue;
      }
      linebuf = readlinebuf;
      truncate_line(linebuf);
    } else {
      finished = 1;
      continue;
    }

    p = linebuf;
    while (isspace(*p))
      p++;

    head = p;
    number = strtoul(p, &top, 10);

    if (!auto_number && (top == p)) { /* direct mode */
      if ((r = nll_line_edit(-1, -1, top, &line)) < 0)
	NLL_ERREXIT(r, top);
      if (line) {
	initialize = 0;
	r = nll_parse(line);
	if (readlinebuf)
	  nll_add_history(nll_line_buffer(line));
	if (r >= 0) {
	  if ((r = nll_start(line, NULL, &retcode)) < 0)
	    finished = 1;
	  if (r > 0)
	    initialize = 1;
	}
	if ((r = nll_line_free(line)) < 0)
	  NLL_ERREXIT(r, top);
	if (initialize) {
	  if ((r = nll_new()) < 0)
	    NLL_ERREXIT(r, top);
	}
      }
    } else { /* program mode */
      if (top == p) {
	number = 0;
	*(--top) = *p ? ' ' : '#';
	number2 = -1;
      } else {
	p = top;
	while (isspace(*p))
	  p++;
	number2 = isdigit(*p) ? strtoul(p, &top, 10) : -1;
      }
      if ((r = nll_line_edit(number, number2, top, &line)) < 0)
	NLL_ERREXIT(r, top);
      if (line)
	nll_parse(line);
      if (readlinebuf) {
	if (line)
	  strcpy(top, nll_line_buffer(line));
	nll_add_history(head);
      }
    }

    if (readlinebuf)
      free(readlinebuf);
  }

  done();

  return retcode;
}
